import contextlib
import functools
import threading
import time
import traceback
import urllib.parse
from time import sleep
from typing import Optional, TypeVar, Set, List, NamedTuple, Callable, TYPE_CHECKING, Dict, Tuple, Any, Generator, Union, Type
from urllib.parse import urlparse

import amqp  # type: ignore
from amqp.exceptions import AMQPError, RecoverableChannelError  # type: ignore

from unipipeline.brokers.uni_broker import UniBroker
from unipipeline.brokers.uni_broker_consumer import UniBrokerConsumer
from unipipeline.definitions.uni_broker_definition import UniBrokerDefinition
from unipipeline.definitions.uni_definition import UniDynamicDefinition
from unipipeline.errors import UniAnswerDelayError, UniMessageRejectError
from unipipeline.message.uni_message import UniMessage
from unipipeline.message_meta.uni_message_meta import UniMessageMeta, UniAnswerParams

if TYPE_CHECKING:
    from unipipeline.modules.uni_mediator import UniMediator

BASIC_PROPERTIES__HEADER__COMPRESSION_KEY = 'compression'

PUBLISHING_RETRYABLE_ERRORS = (AMQPError, *{RecoverableChannelError, *amqp.Connection.recoverable_connection_errors})

T = TypeVar('T')
TFn = TypeVar('TFn', bound=Callable[..., Any])
TMessage = TypeVar('TMessage', bound=UniMessage)


class UniAmqpPyBrokerConfig(UniDynamicDefinition):
    exchange_name: str = "communication"
    answer_exchange_name: str = "communication_answer"
    heartbeat: int = 60  # in seconds
    prefetch: int = 1
    retry_max_count: int = 100
    retry_delay_s: int = 1
    persistent_message: bool = True

    mandatory_publishing: bool = False


class UniAmqpPyBrokerMsgProps(NamedTuple):
    content_type: Optional[str] = None
    content_encoding: Optional[str] = None
    application_headers: Optional[Dict[str, str]] = None
    delivery_mode: Optional[int] = None
    priority: Optional[int] = None
    correlation_id: Optional[str] = None
    reply_to: Optional[str] = None
    expiration: Optional[str] = None
    message_id: Optional[str] = None
    timestamp: Optional[int] = None
    type: Optional[str] = None
    user_id: Optional[str] = None
    app_id: Optional[str] = None
    cluster_id: Optional[str] = None


class UniAmqpPyBrokerConsumer(NamedTuple):
    id: str
    queue: str
    on_message_callback: Callable[[amqp.Channel, 'amqp.Message'], None]
    consumer_tag: str
    prefetch_count: int


class UniAmqpPyBroker(UniBroker[UniAmqpPyBrokerConfig]):
    config_type = UniAmqpPyBrokerConfig

    def _interacted(self) -> float:
        now = time.time()
        with self._lock_interaction:
            self._last_interaction = now
        return now

    @contextlib.contextmanager
    def _interaction(self) -> Generator[float, Any, Any]:
        now = self._interacted()
        yield now
        self._interacted()

    def _get_topic_approximate_messages_count(self, ch: amqp.Channel, topic: str) -> int:
        result = ch.queue_declare(queue=topic, passive=True)
        self.echo.log_debug(f'topic "{topic}" has messages={result.message_count}, consumers={result.consumer_count}')
        return int(result.message_count)

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        with self._channel() as get_ch:
            return self._retry_net_interaction(
                'get_topic_approximate_messages_count',
                lambda has_error: self._get_topic_approximate_messages_count(get_ch(has_error), topic),
                retryable_errors=PUBLISHING_RETRYABLE_ERRORS,
            )

    @classmethod
    def get_connection_uri(cls) -> str:
        raise NotImplementedError(f"cls method get_connection_uri must be implemented for class '{cls.__name__}'")

    @functools.cached_property
    def parsed_connection_uri(self) -> urllib.parse.ParseResult:
        return urlparse(url=self.get_connection_uri())

    __slots__ = (
        '_connection',
        '_consumers',
        '_consuming_enabled',
        '_consumer_in_processing',

        '_interrupted',
        '_initialized_exchanges',
        '_initialized_topics',

        '_heartbeat_enabled',
        '_heartbeat_delay',
        '_heartbeat_thread',

        '_last_interaction',
        '_lock_interaction',

        '_free_channels_lock',
        '_free_channels',
    )

    def __init__(self, mediator: 'UniMediator', definition: UniBrokerDefinition) -> None:
        super().__init__(mediator, definition)

        self._consumers: List[UniAmqpPyBrokerConsumer] = list()

        self._connection = None

        self._consuming_enabled = False
        self._consumer_in_processing = False
        self._interrupted = False

        self._initialized_exchanges: Set[str] = set()
        self._initialized_topics: Set[str] = set()

        self._last_interaction = time.time()
        self._lock_interaction: threading.Lock = threading.Lock()
        self._free_channels_lock: threading.Lock = threading.Lock()
        self._free_channels: List[Tuple[float, amqp.Channel]] = list()

        self._heartbeat_enabled = False
        self._heartbeat_delay = max(self.config.heartbeat / 4, 0.2)
        self._heartbeat_thread: Optional[threading.Thread] = None

    def _close_ch(self, ch: amqp.Channel) -> None:
        ch_id = ch.channel_id
        try:
            with self._interaction():
                ch.close()
        except AMQPError as e:  # noqa
            self.echo.log_warning(f'channel {ch_id} :: closing error :: {str(e)}')
        else:
            self.echo.log_debug(f'channel {ch_id} :: closed successfully')

    def _get_or_create_or_del_free_channel(self) -> Tuple[float, amqp.Channel]:
        now = time.time()
        with self._free_channels_lock:
            for _i in range(len(self._free_channels)):
                c = self._free_channels.pop(-1)
                ch_time, ch = c

                if not ch.is_open or ch.is_closing:
                    continue

                time_left = ch_time - now
                ch_id = ch.channel_id
                if time_left <= 0.:  # REMOVE CHANNEL
                    self._close_ch(ch)
                    continue

                self.echo.log_debug(f'channel {ch_id} :: hold (time left {time_left:0.2f}s)')
                return c

        with self._interaction():
            ch = self.connected_connection.channel()

        self.echo.log_debug(f'channel {ch.channel_id} :: new')

        return now + float(self.config.heartbeat), ch

    @contextlib.contextmanager
    def _channel(self, *, close: bool = False) -> Generator[Callable[[bool], amqp.Channel], Any, Any]:
        c_to_close = []
        current_c = None

        def get_ch(force_new: bool) -> amqp.Channel:
            nonlocal current_c
            if force_new or current_c is None:
                current_c = self._get_or_create_or_del_free_channel()
                c_to_close.append(current_c)
            return current_c[1]

        try:
            yield get_ch
        finally:
            if close:
                for c in c_to_close:
                    self._close_ch(c[1])
            else:
                with self._free_channels_lock:
                    for c in c_to_close:
                        self.echo.log_debug(f'channel {c[1].channel_id} :: release')
                        self._free_channels.append(c)

    @property
    def connected_connection(self) -> amqp.Connection:
        self.connect()
        assert self._connection is not None  # only for mypy
        return self._connection

    def _init_exchange(self, ch: amqp.Channel, exchange: str, *, is_answer: bool) -> None:
        if exchange not in self._initialized_exchanges:
            with self._interaction():
                ch.exchange_declare(
                    exchange=exchange,
                    type="direct",
                    passive=False,
                    durable=True,
                    auto_delete=False,
                )
            self._initialized_exchanges.add(exchange)
            self.echo.log_debug(f'channel {ch.channel_id} :: exchange "{exchange}" was initialized')

    def _just_init_topic(self, ch: amqp.Channel, exchange: str, topic: str, *, is_answer: bool) -> Tuple[str, str]:
        self._init_exchange(ch, exchange, is_answer=is_answer)
        with self._interaction():
            ch.queue_declare(queue=topic, durable=not is_answer, auto_delete=is_answer, exclusive=is_answer, passive=False)
            self.echo.log_debug(f'channel {ch.channel_id} :: queue "{topic}" was initialized')
            ch.queue_bind(queue=topic, exchange=exchange, routing_key=topic)
            self.echo.log_debug(f'channel {ch.channel_id} :: queue "{topic}" was bound to exchange "{exchange}"')
        return exchange, topic

    def _init_answer_topic(self, ch: amqp.Channel, topic: str) -> Tuple[str, str]:
        return self._just_init_topic(ch, self.config.answer_exchange_name, topic, is_answer=True)

    def _init_topic(self, ch: amqp.Channel, topic: str, *, cache_create: bool = True) -> Tuple[str, str]:
        if cache_create and topic in self._initialized_topics:
            return self.config.exchange_name, topic
        res = self._just_init_topic(ch, self.config.exchange_name, topic, is_answer=False)
        self._initialized_topics.add(topic)
        return res

    def initialize(self, topics: Set[str], answer_topic: Set[str]) -> None:
        return

    def stop_consuming(self) -> None:
        self._stop_heartbeat_tick()
        if not self._consuming_enabled:
            self.echo.log_debug('consumption exit skipped')
            return
        self._interrupted = True
        if not self._consumer_in_processing:
            self._consuming_enabled = False
            self.echo.log_debug('consumption stopped')
        else:
            self._consumer_in_processing = False
            self._consuming_enabled = False
            self.echo.log_debug('consumption force stopped')

    def _retry_net_interaction(
        self,
        log_prefix: str,
        fn: Callable[[bool], T],
        *,
        delay: Optional[float] = None,
        retryable_errors: Union[Tuple[Type[Exception], ...], Type[Exception]] = AMQPError
    ) -> T:
        delay = float(self.config.retry_delay_s) if delay is None else delay
        reset_delay = delay * 10
        total_count = max(self.config.retry_max_count, 0)
        retry_count_left = total_count
        has_error = False
        while True:
            last_start_time = time.time()
            try:
                with self._interaction():
                    return fn(has_error)
            except retryable_errors as e:
                retry_count_left -= 1
                has_error = True
                self.echo.log_warning(f'{log_prefix} :: error :: {type(e).__name__} :: {str(e)}')
                self.echo.log_warning(f'{log_prefix} :: retry {total_count - retry_count_left}/{total_count}. delay={delay:0.2f}s')

                sleep(delay)

                if time.time() - last_start_time > reset_delay:
                    has_error = False
                    retry_count_left = total_count
                if not retry_count_left:
                    self.echo.log_error(f'{log_prefix} :: max retry count {total_count} was reached')
                    self.echo.log_error(f'{log_prefix} :: exit on error')
                    self.stop_consuming()
                    raise ConnectionError(f'{log_prefix} :: max retry count {total_count} was reached :: {e}')
            except Exception as e:  # noqa
                self.echo.log_error(f'{log_prefix} :: exit on error')
                self.stop_consuming()
                raise

    def _connect(self) -> None:
        if self._connection is None or not self._connection.connected:
            if self._connection is not None:
                try:
                    self._connection.close()
                except AMQPError:
                    pass
                self._connection = None
            with self._interaction():
                self._connection = amqp.Connection(
                    host=f'{self.parsed_connection_uri.hostname}:{self.parsed_connection_uri.port}',
                    password=self.parsed_connection_uri.password,
                    userid=self.parsed_connection_uri.username,
                    heartbeat=self.config.heartbeat,
                )
                assert self._connection is not None  # only for mypy
                self._connection.connect()
                self._connection.send_heartbeat()
            self.echo.log_debug('connected')

    def connect(self) -> None:  # WITH RETRY
        self._retry_net_interaction(
            'connect',
            lambda _1: self._connect(),
            retryable_errors=PUBLISHING_RETRYABLE_ERRORS,
        )

    def close(self) -> None:
        self._stop_heartbeat_tick()

        if self._connection is None or not self._connection.connected or self._connection.is_closing:
            self._connection = None
            self.echo.log_debug('close :: already closed')
            return

        try:
            with self._interaction():
                self._connection.close()
        except (AMQPError, OSError) as e:  # noqa
            self.echo.log_warning(f'close :: error :: {str(e)}')
        self._connection = None
        self.echo.log_debug('close :: done')

    def add_consumer(self, consumer: UniBrokerConsumer) -> None:
        echo = self.echo.mk_child(f'topic[{consumer.topic}]')
        if self._consuming_enabled:
            raise OverflowError(f'you cannot add consumer dynamically :: tag="{consumer.id}" group_id={consumer.group_id}')

        def consumer_wrapper(ch: amqp.Channel, message: amqp.Message) -> None:
            key = f'consumer "{consumer.id}" :: channel "{ch.channel_id}" :: message "{message.delivery_tag}"'
            self._interacted()
            self._consumer_in_processing = True
            self.echo.log_debug(f'{key} :: received')

            rejected = False
            try:
                get_meta = functools.partial(
                    self.parse_message_body,
                    content=message.body,
                    compression=message.headers.get(BASIC_PROPERTIES__HEADER__COMPRESSION_KEY, None),
                    content_type=message.content_type,
                    unwrapped=consumer.unwrapped,
                )
                consumer.message_handler(get_meta)
            except UniMessageRejectError:
                rejected = True
                self.echo.log_debug(f'{key} :: reject :: started')
                with self._interaction():
                    ch.basic_reject(delivery_tag=message.delivery_tag, requeue=True)
                self.echo.log_debug(f'{key} :: reject :: done')
            except Exception as e: # noqa
                rejected = True
                traceback.print_exc()
                self.echo.log_error(f'{key} :: {str(e)}')
                raise

            if not rejected:
                self.echo.log_debug(f'{key} :: ack :: started')
                with self._interaction():
                    ch.basic_ack(delivery_tag=message.delivery_tag)
                self.echo.log_debug(f'{key} :: ack :: done')

            self._consumer_in_processing = False
            if self._interrupted:
                self.stop_consuming()
                self._close_ch(ch)
            self.echo.log_debug(f'{key} :: processing finished')

        self._consumers.append(UniAmqpPyBrokerConsumer(
            id=consumer.id,
            queue=consumer.topic,
            on_message_callback=consumer_wrapper,
            consumer_tag=consumer.id,
            prefetch_count=consumer.prefetch_count,
        ))

        echo.log_debug(f'added consumer :: tag="{consumer.id}" group_id={consumer.group_id}')

    def _stop_heartbeat_tick(self) -> None:
        if not self._heartbeat_enabled:
            return
        self._heartbeat_enabled = False
        if self._heartbeat_thread is not None:
            self.echo.log_debug('heartbeat :: disable')
            if self._heartbeat_thread.is_alive():
                self._heartbeat_thread.join()
            self._heartbeat_thread = None
        self.echo.log_debug('heartbeat :: disabled')

    def _start_heartbeat_tick(self) -> None:
        self._stop_heartbeat_tick()
        if self._heartbeat_enabled:
            return
        self._heartbeat_enabled = self._heartbeat_delay > 0

        if not self._heartbeat_enabled:
            return

        if self._heartbeat_thread is None:
            self._heartbeat_thread = threading.Thread(
                name=f'broker-{self.definition.name}-system',
                target=self._heartbeat_loop_tick,
                daemon=False,
                kwargs=dict(
                    delay_s=self._heartbeat_delay,
                    heartbeat_delay_threshold=self._heartbeat_delay,
                ),
            )

        if not self._heartbeat_thread.is_alive():
            self._heartbeat_thread.start()

    def _heartbeat_loop_tick(self, delay_s: float, heartbeat_delay_threshold: float) -> None:
        self.echo.log_debug(f'heartbeat :: loop started (delay={delay_s:0.2f}s, threshold={heartbeat_delay_threshold:0.2f}s)')
        step_s = 0.1
        sum_steps = 0.

        while self._heartbeat_enabled:  # is False -> stopping this thread loop
            sleep(step_s)
            sum_steps += step_s
            if sum_steps < delay_s:
                continue
            sum_steps = 0
            with self._lock_interaction:
                current_delay = time.time() - self._last_interaction

            if current_delay < heartbeat_delay_threshold:
                self.echo.log_debug(f'heartbeat :: skipped ({current_delay:0.2f}s > {heartbeat_delay_threshold:0.2f}s)')
                continue

            self.connected_connection.send_heartbeat()
            self.echo.log_debug(f'heartbeat :: tick (since last interaction {current_delay:0.2f}s)')

    def _consuming(self) -> None:
        if not self._consuming_enabled or len(self._consumers) == 0:
            self.echo.log_warning('start_consuming :: has no consumers to start consuming')
            self.stop_consuming()
            return

        self._interrupted = False
        self._consumer_in_processing = False

        self.echo.log_debug('start_consuming')
        self._connect()
        with self._interaction():
            assert self._connection is not None  # only for mypy
            for c in self._consumers:
                ch = self._connection.channel()
                _1, topic = self._init_topic(ch, c.queue, cache_create=False)
                ch.basic_qos(prefetch_count=self.config.prefetch, a_global=False, prefetch_size=0)
                self.echo.log_debug(f'start_consuming :: channel {ch.channel_id} :: set qos (prefetch={c.prefetch_count})')
                ch.basic_consume(queue=topic, callback=functools.partial(c.on_message_callback, ch), consumer_tag=c.consumer_tag)
                self.echo.log_debug(f'start_consuming :: channel {ch.channel_id} :: consumer "{c.consumer_tag}" bound on "{topic}"')

        self._start_heartbeat_tick()
        self.echo.log_debug('start_consuming :: waiting for new messages...')

        while self._consuming_enabled:
            try:
                assert self._connection is not None  # only for mypy
                self._connection.drain_events()
            except Exception as e:  # noqa
                self.echo.log_error(f'consuming loop error :: {e}')
                raise

    def start_consuming(self) -> None:
        self._consuming_enabled = True
        self._retry_net_interaction('consuming', lambda has_error: self._consuming())

    def _has_messages_in_topic(self, ch: amqp.Channel, topic: str, alone: bool) -> bool:
        if alone:
            size = self._get_topic_approximate_messages_count(ch, topic)
            if size > 0:
                self.echo.log_debug(f'sending was skipped, because topic {topic} has messages: {size}>0')
                return True
        return False

    def _publish_ch(self, process_name: str, ch: amqp.Channel, exchange: str, topic: str, meta: UniMessageMeta, props: UniAmqpPyBrokerMsgProps) -> None:
        ch.basic_publish(
            amqp.Message(body=self.serialize_message_body(meta), **props._asdict()),
            exchange=exchange,
            routing_key=topic,
            mandatory=self.config.mandatory_publishing,
            # immediate=self.config.immediate_publishing,
        )
        self.echo.log_debug(f'channel {ch.channel_id} :: {process_name} :: message "{meta.id}" published into "{topic}"')

    def _mk_mesg_props_by_meta(self, meta: UniMessageMeta) -> UniAmqpPyBrokerMsgProps:
        headers = dict()
        if self.definition.compression is not None:
            headers[BASIC_PROPERTIES__HEADER__COMPRESSION_KEY] = self.definition.compression

        ttl_s = meta.real_ttl_s
        expiration = str(ttl_s * 1000) if ttl_s is not None else None

        if meta.need_answer:
            assert meta.answer_params is not None
            return UniAmqpPyBrokerMsgProps(
                content_type=self.definition.content_type,
                content_encoding='utf-8',
                reply_to=self._mk_answer_topic(meta.answer_params),
                correlation_id=str(meta.id),
                delivery_mode=2 if self.config.persistent_message else 0,
                application_headers=headers,
                expiration=expiration,
            )
        return UniAmqpPyBrokerMsgProps(
            content_type=self.definition.content_type,
            content_encoding='utf-8',
            delivery_mode=2 if self.config.persistent_message else 0,
            application_headers=headers,
            expiration=expiration,
        )

    def publish(self, topic: str, meta_list: List[UniMessageMeta], alone: bool = False) -> None:
        with self._channel() as get_ch:
            exchange, topic = self._retry_net_interaction(
                'publish._init_topic',
                lambda has_error: self._init_topic(get_ch(has_error), topic),
                retryable_errors=PUBLISHING_RETRYABLE_ERRORS,
            )

            if self._retry_net_interaction(
                'publish._alone',
                lambda has_error: self._has_messages_in_topic(get_ch(has_error), topic, alone),
                retryable_errors=PUBLISHING_RETRYABLE_ERRORS,
            ):
                return

            for meta in meta_list:
                props = self._mk_mesg_props_by_meta(meta)
                self._retry_net_interaction(
                    'publish._publish_ch',
                    lambda has_error: self._publish_ch('publish', get_ch(has_error), exchange, topic, meta, props),
                    retryable_errors=PUBLISHING_RETRYABLE_ERRORS,
                )

    def rpc_call(self, topic: str, meta: UniMessageMeta, *, alone: bool = False, max_delay_s: int = 1, unwrapped: bool = False) -> Optional[UniMessageMeta]:
        with self._channel(close=True) as get_ch:
            return self._retry_net_interaction(
                'rpc_call',
                lambda has_error: self._rpc_call(get_ch(has_error), topic, meta, alone=alone, max_delay_s=max_delay_s, unwrapped=unwrapped),
                retryable_errors=PUBLISHING_RETRYABLE_ERRORS,
            )

    def _rpc_call(self, ch: amqp.Channel, topic: str, meta: UniMessageMeta, *, alone: bool = False, max_delay_s: int = 1, unwrapped: bool = False) -> Optional[UniMessageMeta]:
        assert meta.answer_params is not None

        answer_exchange, answer_topic = self._init_answer_topic(ch, self._mk_answer_topic(meta.answer_params))

        exchange, topic = self._init_topic(ch, topic)
        if self._has_messages_in_topic(ch, topic, alone):
            return None
        self._publish_ch('rpc_call :: publish', ch, exchange, topic, meta, self._mk_mesg_props_by_meta(meta))

        started = time.time()
        while True:  # TODO: rewrite it to consumer
            msg: Optional[amqp.Message] = ch.basic_get(queue=answer_topic, no_ack=True)
            current_delay = time.time() - started
            if msg is None:
                if current_delay > max_delay_s:
                    raise UniAnswerDelayError(f'channel {ch.channel_id} :: rpc_call :: waiting :: answer for {answer_exchange}->{answer_topic} reached delay limit. {current_delay:0.2f}s > {max_delay_s}s')
                else:
                    self.echo.log_debug(f'channel {ch.channel_id} :: rpc_call :: waiting :: no answer {current_delay:0.2f}s in {answer_exchange}->{answer_topic}')
            else:
                break
            sleep(0.1)

        self.echo.log_debug(f'channel {ch.channel_id} :: rpc_call :: waiting :: found message in "{answer_topic}"!')

        return self.parse_message_body(
            msg.body,
            compression=(msg.headers or dict()).get(BASIC_PROPERTIES__HEADER__COMPRESSION_KEY, None),
            content_type=msg.content_type,
            unwrapped=unwrapped,
        )

    def publish_answer(self, answer_params: UniAnswerParams, meta: UniMessageMeta) -> None:
        headers = dict()

        if self.definition.compression is not None:
            headers[BASIC_PROPERTIES__HEADER__COMPRESSION_KEY] = self.definition.compression

        ttl_s = meta.real_ttl_s
        props = UniAmqpPyBrokerMsgProps(
            content_type=self.definition.content_type,
            content_encoding='utf-8',
            delivery_mode=1,
            application_headers=headers,
            expiration=str(ttl_s * 1000) if ttl_s is not None else None,
        )

        with self._channel() as get_ch:
            self._retry_net_interaction(
                'publish_answer._publish_ch',
                lambda has_error: self._publish_ch('answer :: publish', get_ch(has_error), self.config.answer_exchange_name, self._mk_answer_topic(answer_params), meta, props),
                retryable_errors=PUBLISHING_RETRYABLE_ERRORS,
            )

    def _mk_answer_topic(self, answer_params: UniAnswerParams) -> str:
        return f'{answer_params.topic}.{answer_params.id}'



import functools
from typing import Optional, Any, Dict, List, Set, TYPE_CHECKING, Tuple

from kafka import KafkaProducer, KafkaConsumer  # type: ignore

from unipipeline.brokers.uni_broker import UniBroker
from unipipeline.brokers.uni_broker_consumer import UniBrokerConsumer
from unipipeline.definitions.uni_broker_definition import UniBrokerDefinition
from unipipeline.definitions.uni_dynamic_definition import UniDynamicDefinition
from unipipeline.errors import UniMessageRejectError
from unipipeline.message_meta.uni_message_meta import UniMessageMeta

if TYPE_CHECKING:
    from unipipeline.modules.uni_mediator import UniMediator


class UniKafkaBrokerConfig(UniDynamicDefinition):
    api_version: Tuple[int, ...]
    retry_max_count: int = 100
    retry_delay_s: int = 3


class UniKafkaBroker(UniBroker[UniKafkaBrokerConfig]):
    config_type = UniKafkaBrokerConfig

    def get_boostrap_servers(self) -> List[str]:
        raise NotImplementedError(f'method get_boostrap_server must be implemented for {type(self).__name__}')

    def get_security_conf(self) -> Dict[str, Any]:
        raise NotImplementedError(f'method get_security_conf must be implemented for {type(self).__name__}')

    def __init__(self, mediator: 'UniMediator', definition: UniBrokerDefinition) -> None:
        super().__init__(mediator, definition)

        self._bootstrap_servers = self.get_boostrap_servers()

        self._producer: Optional[KafkaProducer] = None

        self._security_conf: Dict[str, Any] = self.get_security_conf()

        self._consumers: List[UniBrokerConsumer] = list()
        self._kfk_active_consumers: List[KafkaConsumer] = list()

        self._consuming_started = False
        self._interrupted = False
        self._in_processing = False

    def stop_consuming(self) -> None:    # TODO
        self._interrupted = True
        self._end_consuming()

    def _end_consuming(self, force: bool = False) -> None:
        if not self._consuming_started:
            return
        if self._in_processing and not force:
            return
        for kfk_consumer in self._kfk_active_consumers:
            kfk_consumer.close()
        self._kfk_active_consumers.clear()
        self._consuming_started = False
        self.echo.log_info('consumption stopped')

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        return 0  # TODO

    def initialize(self, topics: Set[str], answer_topic: Set[str]) -> None:
        pass  # TODO

    def connect(self) -> None:
        pass

    def close(self) -> None:
        if self._producer is not None:
            self._producer.close()
            self._producer = None
        for kfk_consumer in self._kfk_active_consumers:
            kfk_consumer.close()

    def add_consumer(self, consumer: UniBrokerConsumer) -> None:
        self._consumers.append(consumer)

    def start_consuming(self) -> None:
        echo = self.echo.mk_child('consuming')
        if len(self._consumers) == 0:
            echo.log_warning('has no consumers to start consuming')
            return
        if self._consuming_started:
            echo.log_warning('consuming has already started. ignored')
            return
        self._consuming_started = True
        self._interrupted = False
        self._in_processing = False

        if len(self._consumers) != 1:
            raise OverflowError('invalid consumers number. this type of brokers not supports multiple consumers')

        consumer = self._consumers[0]
        kfk_consumer = KafkaConsumer(
            consumer.topic,
            api_version=self.config.api_version,
            bootstrap_servers=self._bootstrap_servers,
            enable_auto_commit=False,
            group_id=consumer.group_id,
            max_poll_records=1,
            max_poll_interval_ms=3 * 300_000,
            # fetch_max_wait_ms=1000 * 30,  # 30sec
            # max_in_flight_requests_per_connection=3,
            # max_poll_interval_ms=300_000,
            session_timeout_ms=3 * 300_000,  # 1min
        )

        self._kfk_active_consumers.append(kfk_consumer)

        # TODO: retry
        exit_with_error = ''
        for consumer_record in kfk_consumer:
            self._in_processing = True
            echo.log_info(f'consuming message [{consumer_record.offset}]. started')

            get_meta = functools.partial(
                self.parse_message_body,
                content=consumer_record.value,
                compression=self.definition.compression,
                content_type=self.definition.content_type,
                unwrapped=consumer.unwrapped,
            )

            rejected = False
            try:
                consumer.message_handler(get_meta)
            except UniMessageRejectError as e:
                rejected = True
                echo.log_warning(f'consuming message [{consumer_record.offset}]. reject {type(e).__name__}. {e}')
            except Exception as e:  # noqa
                echo.log_error(f'consuming message [{consumer_record.offset}]. error {type(e).__name__}. {e}')
                raise

            self._in_processing = False
            if not rejected:
                try:
                    kfk_consumer.commit()
                except Exception as e:  # noqa
                    exit_with_error = f'consuming message [{consumer_record.offset}]. error {type(e).__name__}. {e}'
                    break
                else:
                    echo.log_info(f'consuming message [{consumer_record.offset}]. ok')
            if self._interrupted:
                break

        self._end_consuming(True)

        if exit_with_error:
            echo.exit_with_error(exit_with_error)

    def _connect_producer(self) -> None:
        if self._producer is not None:
            if self._producer._closed:
                self._producer.close()
                self._producer = None
            else:
                return

        # TODO: change default connection as producer yo abstract connection to kafka server
        self._producer = KafkaProducer(
            bootstrap_servers=self._bootstrap_servers,
            api_version=self.config.api_version,
            retries=self.config.retry_max_count,
            acks=1,
            **self._security_conf,
        )

        if not self._producer.bootstrap_connected():
            raise ConnectionError()

        self.echo.log_info('connected')

    def _get_producer(self) -> KafkaProducer:
        self._connect_producer()
        assert self._producer is not None
        return self._producer

    def publish(self, topic: str, meta_list: List[UniMessageMeta], alone: bool = False) -> None:
        # TODO: alone
        # TODO: ttl
        # TODO: retry
        self.echo.log_debug(f'publishing the messages: {meta_list}')

        p = self._get_producer()

        for meta in meta_list:
            p.send(
                topic=topic,
                value=self.serialize_message_body(meta),
                key=str(meta.id).encode('utf8')
            )
        p.flush()



from datetime import datetime
from typing import TypeVar, Optional, Generic, Type
from uuid import UUID

from unipipeline.errors import UniMessagePayloadParsingError, UniMessageRejectError
from unipipeline.message.uni_message import UniMessage
from unipipeline.message_meta.uni_message_meta import UniMessageMeta

TInputMsgPayload = TypeVar('TInputMsgPayload', bound=UniMessage)
TAnswerMsgPayload = TypeVar('TAnswerMsgPayload', bound=Optional[UniMessage])


class UniWorkerConsumerMessage(Generic[TInputMsgPayload]):
    def __init__(self, message_input: Type[TInputMsgPayload], meta: UniMessageMeta) -> None:
        self._meta = meta
        self._message_input_payload_type = message_input
        self._message_payload_cache: Optional[TInputMsgPayload] = None
        self._acknowledged_or_rejected = False

    def reject(self) -> None:
        raise UniMessageRejectError()

    @property
    def id(self) -> UUID:
        return self._meta.id

    @property
    def date_created(self) -> datetime:
        return self._meta.date_created

    @property
    def worker_creator(self) -> Optional[str]:
        return self._meta.worker_creator

    @property
    def payload(self) -> TInputMsgPayload:
        if isinstance(self._message_payload_cache, UniMessage):
            return self._message_payload_cache  # type: ignore

        try:
            self._message_payload_cache = self._message_input_payload_type(**self._meta.payload)
        except Exception as e:  # noqa
            raise UniMessagePayloadParsingError(str(e))

        return self._message_payload_cache



from typing import Optional, Any, Dict
from uuid import UUID

from pydantic import root_validator, validator

from unipipeline.definitions.uni_definition import UniDefinition
from unipipeline.definitions.uni_worker_definition import UniWorkerDefinition


class UniCronTaskDefinition(UniDefinition):
    id: UUID
    name: str
    worker: UniWorkerDefinition
    when: Optional[str]
    every_sec: Optional[int]
    alone: bool

    @validator('every_sec')
    def validate_every_sec(cls, v: Optional[int], values: Dict[str, Any], **kwargs: Any) -> Optional[int]:
        if v is None:
            return v
        if not isinstance(v, int):
            TypeError('Invalid type. must be int')
        if v <= 1:
            raise ValueError('Must be > 1')
        if 60 % v:
            raise ValueError('60 must be a multiple of N')
        return v

    @root_validator()
    def validate_all(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        when = values.get('when', None)
        every_sec = values.get('every_sec', None)
        if not ((when is not None) ^ (every_sec is not None)):
            raise ValueError(f'cron "{values["name"]}" has property conflict in (when, every_sec). one from it must be None')
        return values



from example.messages.ender_after_answer_message import EnderAfterAnswerMessage
from example.messages.ender_after_input_message import EnderAfterInputMessage
from unipipeline.worker.uni_worker import UniWorker
from unipipeline.worker.uni_worker_consumer_message import UniWorkerConsumerMessage


class EnderAfterInputWorker(UniWorker[EnderAfterInputMessage, EnderAfterAnswerMessage]):
    def handle_message(self, msg: UniWorkerConsumerMessage[EnderAfterInputMessage]) -> EnderAfterAnswerMessage:
        # print(f'>>>>> {msg.payload}')  # noqa
        return EnderAfterAnswerMessage(
            value=f'EnderAfterInputWorker answer on >>> {msg.payload.value}'
        )



import argparse

import logging
from sys import stdout

logging.getLogger('amqp').setLevel(logging.DEBUG)
logging.getLogger('amqp.connection.Connection.heartbeat_tick').setLevel(logging.DEBUG)
logging.getLogger('amqp.connection.Connection').setLevel(logging.DEBUG)
logging.getLogger('amqp.abstract_channel').setLevel(logging.DEBUG)
logging.getLogger('amqp.basic_message').setLevel(logging.DEBUG)
logging.getLogger('amqp.channel').setLevel(logging.DEBUG)
logging.getLogger('amqp.connection').setLevel(logging.DEBUG)
logging.getLogger('amqp.exceptions').setLevel(logging.DEBUG)
logging.getLogger('amqp.method_framing').setLevel(logging.DEBUG)
logging.getLogger('amqp.platform').setLevel(logging.DEBUG)
logging.getLogger('amqp.protocol').setLevel(logging.DEBUG)
logging.getLogger('amqp.sasl').setLevel(logging.DEBUG)
logging.getLogger('amqp.serialization').setLevel(logging.DEBUG)
logging.getLogger('amqp.spec').setLevel(logging.DEBUG)
logging.getLogger('amqp.transport').setLevel(logging.DEBUG)
logging.getLogger('amqp.utils').setLevel(logging.DEBUG)
logging.getLogger('pamqp').setLevel(logging.DEBUG)
logging.getLogger('pamqp.frame').setLevel(logging.DEBUG)
logging.getLogger('socket').setLevel(logging.DEBUG)

logging.basicConfig(level=logging.DEBUG, handlers=[logging.StreamHandler(stream=stdout)])

parser = argparse.ArgumentParser()
parser.add_argument('--type', dest='type', choices=('mem', 'kafka', 'rmq'), type=str)
parser.add_argument('--count', dest='produce_count', type=int, required=False, default=3)
parser.add_argument('--batch-count', dest='produce_batch_count', type=int, required=False, default=1)
parser.add_argument('--delay', dest='delay', type=int, required=False, default=1)
parser.add_argument('--worker', dest='worker', type=str, required=False, default=None)

args = parser.parse_args()



from example.messages.input_message import InputMessage
from unipipeline.worker.uni_worker import UniWorker
from unipipeline.worker.uni_worker_consumer_message import UniWorkerConsumerMessage


class AutoRetryWorker(UniWorker[InputMessage, None]):
    def handle_message(self, msg: UniWorkerConsumerMessage[InputMessage]) -> None:
        raise RuntimeError('some critical runtime error. like db deadlock')



from urllib.parse import urlparse

import psycopg2  # type: ignore

from unipipeline.waiting.uni_wating import UniWaiting


class UniPostgresWaiting(UniWaiting):
    close_after_success: bool = True

    def get_connection_uri(self) -> str:
        raise NotImplementedError(f'method get_connection_uri was not specified for {type(self).__name__}')

    def try_to_connect(self) -> None:
        parsed_db_uri = urlparse(self.get_connection_uri())
        connection = psycopg2.connect(
            dbname=parsed_db_uri.path.strip('/'),
            user=parsed_db_uri.username,
            password=parsed_db_uri.password,
            host=parsed_db_uri.hostname,
            port=parsed_db_uri.port,
        )

        if self.close_after_success:
            connection.close()



import functools
import time
from time import sleep
from typing import Optional, TypeVar, Set, List, NamedTuple, Callable, TYPE_CHECKING
from urllib.parse import urlparse

from pika import ConnectionParameters, PlainCredentials, BlockingConnection, BasicProperties, spec  # type: ignore
from pika.adapters.blocking_connection import BlockingChannel  # type: ignore
from pika.exceptions import AMQPConnectionError, AMQPError, ConnectionClosedByBroker, AMQPChannelError  # type: ignore

from unipipeline.brokers.uni_broker import UniBroker
from unipipeline.brokers.uni_broker_consumer import UniBrokerConsumer
from unipipeline.definitions.uni_broker_definition import UniBrokerDefinition
from unipipeline.definitions.uni_definition import UniDynamicDefinition
from unipipeline.errors import UniAnswerDelayError, UniMessageRejectError
from unipipeline.message.uni_message import UniMessage
from unipipeline.message_meta.uni_message_meta import UniMessageMeta, UniAnswerParams
from unipipeline.utils.uni_echo import UniEcho

if TYPE_CHECKING:
    from unipipeline.modules.uni_mediator import UniMediator

BASIC_PROPERTIES__HEADER__COMPRESSION_KEY = 'compression'

CONNECTION_ERRORS = (ConnectionClosedByBroker, AMQPConnectionError)


TMessage = TypeVar('TMessage', bound=UniMessage)

T = TypeVar('T')


class UniAmqpPikaBrokerConfig(UniDynamicDefinition):
    exchange_name: str = "communication"
    answer_exchange_name: str = "communication_answer"
    heartbeat: int = 600
    blocked_connection_timeout: int = 300
    prefetch: int = 1
    retry_max_count: int = 100
    retry_delay_s: int = 3
    socket_timeout: int = 300
    stack_timeout: int = 400
    persistent_message: bool = True


class UniAmqpPikaBrokerConsumer(NamedTuple):
    queue: str
    on_message_callback: Callable[[BlockingChannel, spec.Basic.Deliver, BasicProperties, bytes], None]
    consumer_tag: str
    prefetch_count: int


class AmqpPikaChannelObj:
    def __init__(self, name: str, ttl: int, broker: 'UniAmqpPikaBroker') -> None:
        self._name = name
        self._broker = broker
        self._ch: Optional[BlockingChannel] = None
        self._ttl = ttl
        self._ch_time: float = 0

    def close_channel(self) -> None:
        if self._ch is None or self._ch.is_closed:
            return
        self._ch.close()
        self._broker.echo.log_info(f'channel {self._name} closed')

    def get_channel(self, *, force_recreate: bool = False, force_current: bool = False) -> BlockingChannel:
        if force_recreate and not force_current:
            self.close_channel()
            return self.get_channel(force_recreate=False)

        if self._ch is None or self._ch.is_closed:
            self._broker.connect()
            conn = self._broker._connection
            assert conn is not None
            self._ch = conn.channel()
            self._ch_time = time.time()
            self._broker.echo.log_info(f'channel {self._name} established')
        else:
            if not force_current and (time.time() - self._ch_time > self._ttl):
                return self.get_channel(force_recreate=True)
            self._broker.echo.log_debug(f'channel {self._name} reused')
        return self._ch


class UniAmqpPikaBroker(UniBroker[UniAmqpPikaBrokerConfig]):
    config_type = UniAmqpPikaBrokerConfig

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        res = self._ch_stat.get_channel().queue_declare(queue=topic, passive=True)
        return int(res.method.message_count)

    @classmethod
    def get_connection_uri(cls) -> str:
        raise NotImplementedError(f"cls method get_connection_uri must be implemented for class '{cls.__name__}'")

    def __init__(self, mediator: 'UniMediator', definition: UniBrokerDefinition) -> None:
        super().__init__(mediator, definition)

        broker_url = self.get_connection_uri()

        url_params_pr = urlparse(url=broker_url)

        self._params = ConnectionParameters(
            heartbeat=self.config.heartbeat,
            blocked_connection_timeout=self.config.blocked_connection_timeout,
            socket_timeout=self.config.socket_timeout,
            stack_timeout=self.config.stack_timeout,
            retry_delay=self.definition.retry_delay_s,
            host=url_params_pr.hostname,
            port=url_params_pr.port,
            credentials=PlainCredentials(url_params_pr.username, url_params_pr.password, erase_on_connect=False),
        )

        self._consumers: List[UniAmqpPikaBrokerConsumer] = list()

        self._connection: Optional[BlockingConnection] = None

        self._ch_initializer = AmqpPikaChannelObj('initializer', int(self.config.heartbeat / 2), self)
        self._ch_stat = AmqpPikaChannelObj('stat', int(self.config.heartbeat / 2), self)
        self._ch_publisher = AmqpPikaChannelObj('publisher', int(self.config.heartbeat / 2), self)
        self._ch_answ_publisher = AmqpPikaChannelObj('answer_publisher', int(self.config.heartbeat / 2), self)
        self._ch_consumer = AmqpPikaChannelObj('consumer', int(self.config.heartbeat / 2), self)
        self._ch_answ_consumer = AmqpPikaChannelObj('answer_consumer', int(self.config.heartbeat / 2), self)

        self._consuming_started = False
        self._in_processing = False
        self._interrupted = False

        self._initialized_exchanges: Set[str] = set()
        self._initialized_topics: Set[str] = set()

    def _init_exchange(self, ch: BlockingChannel, exchange: str) -> None:
        if exchange in self._initialized_topics:
            return
        self._initialized_exchanges.add(exchange)
        ch.exchange_declare(
            exchange=exchange,
            exchange_type="direct",
            passive=False,
            durable=True,
            auto_delete=False,
        )
        self.echo.log_info(f'exchange "{exchange}" initialized')

    def _init_topic(self, ch: BlockingChannel, exchange: str, topic: str) -> str:
        q = f'{exchange}->{topic}'
        if q in self._initialized_topics:
            return topic
        self._initialized_topics.add(q)

        self._init_exchange(ch, exchange)

        if exchange == self.config.exchange_name:
            ch.queue_declare(queue=topic, durable=True, auto_delete=False, passive=False)
        elif exchange == self.config.answer_exchange_name:
            ch.queue_declare(queue=topic, durable=False, auto_delete=True, exclusive=True, passive=False)
        else:
            raise ValueError(f'invalid exchange name "{exchange}"')

        ch.queue_bind(queue=topic, exchange=self.config.exchange_name, routing_key=topic)
        self.echo.log_info(f'queue "{q}" initialized')
        return topic

    def initialize(self, topics: Set[str], answer_topic: Set[str]) -> None:
        return

    def stop_consuming(self) -> None:
        self._end_consuming()

    def _end_consuming(self) -> None:
        if not self._consuming_started:
            return
        self._interrupted = True
        if not self._in_processing:
            self._ch_consumer.get_channel(force_current=True).stop_consuming()
            self.close()
            self._consuming_started = False
            self.echo.log_info('consumption stopped')

    def connect(self) -> None:
        try:
            if self._connection is None or self._connection.is_closed:
                self._connection = BlockingConnection(self._params)
                self.echo.log_info('connected')
        except (*CONNECTION_ERRORS, AMQPError) as e:
            raise ConnectionError(str(e))

    def close(self) -> None:
        if self._connection is None:
            return

        if self._connection.is_closed:
            self._connection = None
            return

        try:
            self._connection.close()
            self._connection = None
        except AMQPError:
            pass

    def add_consumer(self, consumer: UniBrokerConsumer) -> None:
        echo = self.echo.mk_child(f'topic[{consumer.topic}]')
        if self._consuming_started:
            echo.exit_with_error(f'you cannot add consumer dynamically :: tag="{consumer.id}" group_id={consumer.group_id}')

        def consumer_wrapper(channel: BlockingChannel, method_frame: spec.Basic.Deliver, properties: BasicProperties, body: bytes) -> None:
            self._in_processing = True

            get_meta = functools.partial(
                self.parse_message_body,
                content=body,
                compression=properties.headers.get(BASIC_PROPERTIES__HEADER__COMPRESSION_KEY, None),
                content_type=properties.content_type,
                unwrapped=consumer.unwrapped,
            )

            rejected = False
            try:
                consumer.message_handler(get_meta)
            except UniMessageRejectError:
                channel.basic_reject(delivery_tag=method_frame.delivery_tag, requeue=True)
                rejected = True

            if not rejected:
                channel.basic_ack(delivery_tag=method_frame.delivery_tag)

            self._in_processing = False
            if self._interrupted:
                self._end_consuming()

        self._consumers.append(UniAmqpPikaBrokerConsumer(
            queue=consumer.topic,
            on_message_callback=consumer_wrapper,
            consumer_tag=consumer.id,
            prefetch_count=consumer.prefetch_count,
        ))

        echo.log_info(f'added consumer :: tag="{consumer.id}" group_id={consumer.group_id}')

    def _start_consuming(self) -> None:
        echo = self.echo.mk_child('consuming')
        if len(self._consumers) == 0:
            echo.log_warning('has no consumers to start consuming')
            return
        ch = self._ch_consumer.get_channel()
        prefetch_count: int = self.config.prefetch
        for c in self._consumers:
            topic = self._init_topic(ch, self.config.exchange_name, c.queue)
            ch.basic_consume(queue=topic, on_message_callback=c.on_message_callback, consumer_tag=c.consumer_tag)
            echo.log_debug(f'added consumer {c.consumer_tag} on {self.config.exchange_name}->{topic}')
            prefetch_count = max(prefetch_count, c.prefetch_count)
        echo.log_info(f'consumers count is {len(self._consumers)}')
        ch.basic_qos(prefetch_count=prefetch_count)
        ch.start_consuming()  # blocking operation

    def _retry_run(self, echo: UniEcho, fn: Callable[[], T]) -> T:
        retry_counter = 0
        max_retries = max(self.config.retry_max_count, 1)
        retry_threshold_s = self.config.retry_delay_s * max_retries
        while True:
            start = time.time()
            try:
                return fn()
            except AMQPChannelError as e:
                echo.log_warning(f"Caught a channel error: {e}, stopping...")
                raise
            except CONNECTION_ERRORS as e:
                echo.log_error(f'connection closed {e}')
                if int(time.time() - start) >= retry_threshold_s:
                    retry_counter = 0
                if retry_counter >= max_retries:
                    raise ConnectionError()
                retry_counter += 1
                sleep(self.config.retry_delay_s)
                echo.log_warning(f'retry {retry_counter}/{max_retries} :: {e}')

    def start_consuming(self) -> None:
        echo = self.echo.mk_child('consuming')
        if self._consuming_started:
            echo.log_warning('consuming has already started. ignored')
            return
        self._consuming_started = True
        self._interrupted = False
        self._in_processing = False

        self._retry_run(echo, self._start_consuming)

    def _publish(self, ch: BlockingChannel, exchange: str, topic: str, meta: UniMessageMeta, props: BasicProperties, alone: bool = False) -> None:
        self.echo.log_debug(f'message start publishing to {exchange}->{topic}')
        if alone:
            size = self.get_topic_approximate_messages_count(topic)
            if size != 0:
                self.echo.log_info(f'sending was skipped, because topic {topic} has messages: {size}>0')
                return
        topic = self._init_topic(ch, exchange, topic)
        ch.basic_publish(
            exchange=exchange,
            routing_key=topic,
            body=self.serialize_message_body(meta),
            properties=props
        )
        self.echo.log_debug(f'message published to {exchange}->{topic}')

    def publish(self, topic: str, meta_list: List[UniMessageMeta], alone: bool = False) -> None:
        ch = self._ch_publisher.get_channel()
        echo = self.echo.mk_child('publish')
        for meta in meta_list:  # TODO: package sending
            headers = dict()
            if self.definition.compression is not None:
                headers[BASIC_PROPERTIES__HEADER__COMPRESSION_KEY] = self.definition.compression
            if meta.ttl_s:
                headers['x-message-ttl'] = str(meta.ttl_s * 1000)

            if meta.need_answer:
                assert meta.answer_params is not None
                props = BasicProperties(
                    content_type=self.definition.content_type,
                    content_encoding='utf-8',
                    reply_to=self._mk_answer_topic(meta.answer_params),
                    correlation_id=str(meta.id),
                    delivery_mode=2 if self.config.persistent_message else 0,
                    headers=headers,
                )
            else:
                props = BasicProperties(
                    content_type=self.definition.content_type,
                    content_encoding='utf-8',
                    delivery_mode=2 if self.config.persistent_message else 0,
                    headers=headers,
                )
            self._retry_run(echo, functools.partial(self._publish, ch=ch, exchange=self.config.exchange_name, topic=topic, meta=meta, props=props, alone=alone))
        self.echo.log_info(f'{list(meta_list)} messages published to {self.config.exchange_name}->{topic}')

    def _get_answ(self, answer_params: UniAnswerParams, max_delay_s: int, unwrapped: bool) -> UniMessageMeta:
        ch = self._ch_answ_consumer.get_channel(force_recreate=True)
        topic = self._init_topic(ch, self.config.answer_exchange_name, self._mk_answer_topic(answer_params))
        exchange = self.config.exchange_name

        started = time.time()
        while True:
            (method, properties, body) = ch.basic_get(queue=topic, auto_ack=True)

            if method is None:
                if (time.time() - started) > max_delay_s:
                    raise UniAnswerDelayError(f'answer for {self.config.answer_exchange_name}->{topic} reached delay limit {max_delay_s} seconds')
                self.echo.log_debug(f'no answer {int(time.time() - started + 1)}s in {exchange}->{topic}')
                sleep(0.33)
                continue

            self.echo.log_debug(f'took answer from {self.config.answer_exchange_name}->{topic}')
            return self.parse_message_body(
                body,
                compression=properties.headers.get(BASIC_PROPERTIES__HEADER__COMPRESSION_KEY, None),
                content_type=properties.content_type,
                unwrapped=unwrapped,
            )

    def get_answer(self, answer_params: UniAnswerParams, max_delay_s: int, unwrapped: bool) -> UniMessageMeta:
        echo = self.echo.mk_child('get_answer')
        return self._retry_run(echo, functools.partial(self._get_answ, answer_params=answer_params, max_delay_s=max_delay_s, unwrapped=unwrapped))

    def publish_answer(self, answer_params: UniAnswerParams, meta: UniMessageMeta) -> None:
        echo = self.echo.mk_child('publish_answer')
        props = BasicProperties(
            content_type=self.definition.content_type,
            content_encoding='utf-8',
            delivery_mode=1,
        )
        ch = self._ch_answ_publisher.get_channel()
        self._retry_run(echo, functools.partial(self._publish, ch=ch, exchange=self.config.answer_exchange_name, topic=self._mk_answer_topic(answer_params), meta=meta, props=props))

    def _mk_answer_topic(self, answer_params: UniAnswerParams) -> str:
        return f'{answer_params.topic}.{answer_params.id}'
